﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml.Linq;
using System.Xml.XPath;
using System.Xml;
using System.IO;
using ICSharpCode.SharpZipLib.Zip;

namespace BAL
{
    public delegate string ProcessFormattingDelegate();

    public class ExcelFormattingAdjustor
    {
        private Regex reg = new Regex("(\\d+\\,\\d+)|(\\d+)+");

        #region Const strings for Excel sheets formatting replacing

        private const string NAMESPACE = "http://schemas.openxmlformats.org/spreadsheetml/2006/main";
        private const string BALANCESHEETS_A10 = "<r>"
                                               + "  <t xml:space=\"preserve\">  Short-term investments (including securities loaned\r\n    of </t>"
                                               + "</r>"
                                               + "<r>"
                                               + "  <rPr>"
                                               + "    <b/>"
                                               + "    <sz val=\"8\"/>"
                                               + "    <color rgb=\"FF404040\"/>"
                                               + "    <rFont val=\"Verdana\"/>"
                                               + "    <family val=\"2\"/>"
                                               + "  </rPr>"
                                               + "  <t>${0}</t>"
                                               + "</r>"
                                               + "<r>"
                                               + "  <rPr>"
                                               + "    <sz val=\"8\"/>"
                                               + "    <color rgb=\"FF404040\"/>"
                                               + "    <rFont val=\"Verdana\"/>"
                                               + "    <family val=\"2\"/>"
                                               + "  </rPr>"
                                               + "  <t xml:space=\"preserve\"> and ${1})</t>"
                                               + "</r>";
        private const string BALANCESHEETS_A12 = "<r>"
                                               + "  <t xml:space=\"preserve\">  Accounts receivable, net of allowance for doubtful\r\n    accounts of </t>"
                                               + "</r>"
                                               + "<r>"
                                               + "  <rPr>"
                                               + "    <b/>"
                                               + "    <sz val=\"8\"/>"
                                               + "    <color rgb=\"FF404040\"/>"
                                               + "    <rFont val=\"Verdana\"/>"
                                               + "    <family val=\"2\"/>"
                                               + "  </rPr>"
                                               + "  <t>${0}</t>"
                                               + "</r>"
                                               + "<r>"
                                               + "  <rPr>"
                                               + "    <sz val=\"8\"/>"
                                               + "    <color rgb=\"FF404040\"/>"
                                               + "    <rFont val=\"Verdana\"/>"
                                               + "    <family val=\"2\"/>"
                                               + "  </rPr>"
                                               + "  <t xml:space=\"preserve\"> and ${1}</t>"
                                               + "</r>";
        private const string BALANCESHEETS_A17 = "<r>"
                                               + "  <t xml:space=\"preserve\">Property and equipment, net of accumulated\r\n  depreciation of </t>"
                                               + "</r>"
                                               + "<r>"
                                               + "  <rPr>"
                                               + "    <b/>"
                                               + "    <sz val=\"8\"/>"
                                               + "    <color rgb=\"FF404040\"/>"
                                               + "    <rFont val=\"Verdana\"/>"
                                               + "    <family val=\"2\"/>"
                                               + "  </rPr>"
                                               + "  <t>${0}</t>"
                                               + "</r>"
                                               + "<r>"
                                               + "  <rPr>"
                                               + "    <sz val=\"8\"/>"
                                               + "    <color rgb=\"FF404040\"/>"
                                               + "    <rFont val=\"Verdana\"/>"
                                               + "    <family val=\"2\"/>"
                                               + "  </rPr>"
                                               + "  <t xml:space=\"preserve\"> and ${1}</t>"
                                               + "</r>";
        private const string BALANCESHEETS_A40 = "<r>"
                                               + "  <t xml:space=\"preserve\">  Common stock and paid-in capital - shares authorized\r\n    {0}; outstanding </t>"
                                               + "</r>"
                                               + "<r>"
                                               + "  <rPr>"
                                               + "    <b/>"
                                               + "    <sz val=\"8\"/>"
                                               + "    <color rgb=\"FF404040\"/>"
                                               + "    <rFont val=\"Verdana\"/>"
                                               + "    <family val=\"2\"/>"
                                               + "  </rPr>"
                                               + "  <t>{1}</t>"
                                               + "</r>"
                                               + "<r>"
                                               + "  <rPr>"
                                               + "    <sz val=\"8\"/>"
                                               + "    <color rgb=\"FF404040\"/>"
                                               + "    <rFont val=\"Verdana\"/>"
                                               + "    <family val=\"2\"/>"
                                               + "  </rPr>"
                                               + "  <t xml:space=\"preserve\"> and {2}</t>"
                                               + "</r>";
        private const string BALANCESHEETS_A41 = "<r>"/*-----------------Need special handling--------------------*/
                                               + "  <t xml:space=\"preserve\">  {0}, including accumulated other\r\n    comprehensive income of </t>"
                                               + "</r>"
                                               + "<r>"
                                               + "  <rPr>"
                                               + "    <b/>"
                                               + "    <sz val=\"8\"/>"
                                               + "    <color rgb=\"FF404040\"/>"
                                               + "    <rFont val=\"Verdana\"/>"
                                               + "    <family val=\"2\"/>"
                                               + "  </rPr>"
                                               + "  <t>${1}</t>"
                                               + "</r>"
                                               + "<r>"
                                               + "  <rPr>"
                                               + "    <sz val=\"8\"/>"
                                               + "    <color rgb=\"FF404040\"/>"
                                               + "    <rFont val=\"Verdana\"/>"
                                               + "    <family val=\"2\"/>"
                                               + "  </rPr>"
                                               + "  <t xml:space=\"preserve\"> and ${2}</t>"
                                               + "</r>";
        private const string BALANCESHEETS_B6 = "<t>{0},\r\n{1}</t>";/*-----------------Need special handling--------------------*/
        private const string BALANCESHEETS_D6 = "<r>"/*-----------------Need special handling--------------------*/
                                              + "  <t>{0},\r\n{1}</t>"
                                              + "</r>"
                                              + "<r>"
                                              + "  <rPr>"
                                              + "    <b/>"
                                              + "    <vertAlign val=\"superscript\"/>"
                                              + "    <sz val=\"8\"/>"
                                              + "    <color rgb=\"FF404040\"/>"
                                              + "    <rFont val=\"Verdana\"/>"
                                              + "    <family val=\"2\"/>"
                                              + "  </rPr>"
                                              + "  <t>(1)</t>"
                                              + "</r>";
        private const string CASHFLOWS_A14 = "<t xml:space=\"preserve\">  {0} on\r\n    investments and derivatives</t>";/*-----------------Need special handling--------------------*/
        private const string CASHFLOWS_A29 = "<t>{0},\r\n  maturities of 90 days or less, net</t>";/*-----------------Need special handling--------------------*/
        private const string CASHFLOWS_A37 = "<t xml:space=\"preserve\">        {0}\r\n          financing</t>";/*-----------------Need special handling--------------------*/
        private const string CASHFLOWS_A45 = "<t xml:space=\"preserve\">        {0} investing</t>"/*-----------------Need special handling--------------------*/
                                           + "<phoneticPr fontId=\"41\" type=\"noConversion\"/>";

        #endregion

        private List<ExcelFormattingSetting> ExcelFormattingSettings;

        public ExcelFormattingAdjustor()
        {
            ExcelFormattingSettings = new List<ExcelFormattingSetting>();

            ExcelFormattingSettings.Add(new ExcelFormattingSetting() { FormattingText = BALANCESHEETS_A10, ExcelPositionString = "A10", SheetName = "xl/worksheets/sheet3.xml" });
            ExcelFormattingSettings.Add(new ExcelFormattingSetting() { FormattingText = BALANCESHEETS_A12, ExcelPositionString = "A12", SheetName = "xl/worksheets/sheet3.xml" });
            ExcelFormattingSettings.Add(new ExcelFormattingSetting() { FormattingText = BALANCESHEETS_A17, ExcelPositionString = "A17", SheetName = "xl/worksheets/sheet3.xml" });
            ExcelFormattingSettings.Add(new ExcelFormattingSetting() { FormattingText = BALANCESHEETS_A40, ExcelPositionString = "A40", SheetName = "xl/worksheets/sheet3.xml" });

            #region Special handling for BalanceSheets A41
            ExcelFormattingSetting settingBalanceA41 = new ExcelFormattingSetting(BALANCESHEETS_A41, "A41", "xl/worksheets/sheet3.xml");
            settingBalanceA41.processFormatting = () => {
                string[] parts = settingBalanceA41.OriginalText.Split(',');
                if (parts.Length > 1)
                {
                    MatchCollection matches = reg.Matches(parts[1]);
                    if (matches.Count > 0)
                    {
                        string[] arr = new string[matches.Count + 1];
                        arr[0] = parts[0].Trim();
                        for (int i = 1; i < matches.Count; i++)
                        {
                            arr[i] = matches[i].ToString();
                        }
                        return string.Format(BALANCESHEETS_A41, arr);
                    }
                }
                
                return string.Empty; 
            };
            ExcelFormattingSettings.Add(settingBalanceA41);
            #endregion

            #region Special handling for BalanceSheets B6
            ExcelFormattingSetting settingBalanceB6 = new ExcelFormattingSetting(BALANCESHEETS_B6, "B6", "xl/worksheets/sheet3.xml");
            settingBalanceB6.processFormatting = () =>
            {
                string[] parts = settingBalanceB6.OriginalText.Split(',');
                if (parts.Length > 1)
                {
                    return string.Format(BALANCESHEETS_B6, parts);
                }

                return string.Empty;
            };
            ExcelFormattingSettings.Add(settingBalanceB6);
            #endregion

            #region Special handling for BalanceSheets D6
            ExcelFormattingSetting settingBalanceD6 = new ExcelFormattingSetting(BALANCESHEETS_D6, "D6", "xl/worksheets/sheet3.xml");
            settingBalanceD6.processFormatting = () =>
            {
                string[] parts = settingBalanceD6.OriginalText.Split(',');
                if (parts.Length > 1)
                {
                    return string.Format(BALANCESHEETS_D6, parts);
                }

                return string.Empty;
            };
            ExcelFormattingSettings.Add(settingBalanceD6);
            #endregion

            #region Special handling for CashFlows A14
            ExcelFormattingSetting settingCashA14 = new ExcelFormattingSetting(CASHFLOWS_A14, "A14", "xl/worksheets/sheet4.xml");
            settingCashA14.processFormatting = () =>
            {
                string[] parts = settingCashA14.OriginalText.Split(new string[]{" on "}, StringSplitOptions.None);
                if (parts.Length > 1)
                {
                    return string.Format(CASHFLOWS_A14, parts[0].Trim());
                }

                return string.Empty;
            };
            ExcelFormattingSettings.Add(settingCashA14);
            #endregion

            #region Special handling for CashFlows A29
            ExcelFormattingSetting settingCashA29 = new ExcelFormattingSetting(CASHFLOWS_A29, "A29", "xl/worksheets/sheet4.xml");
            settingCashA29.processFormatting = () =>
            {
                string[] parts = settingCashA29.OriginalText.Split(',');
                if (parts.Length > 1)
                {
                    return string.Format(CASHFLOWS_A29, parts[0].Trim());
                }

                return string.Empty;
            };
            ExcelFormattingSettings.Add(settingCashA29);
            #endregion

            #region Special handling for CashFlows A37
            ExcelFormattingSetting settingCashA37 = new ExcelFormattingSetting(CASHFLOWS_A37, "A37", "xl/worksheets/sheet4.xml");
            settingCashA37.processFormatting = () =>
            {
                string part = settingCashA37.OriginalText.Substring(0, settingCashA37.OriginalText.Length - 9);
                return string.Format(CASHFLOWS_A37, part.Trim());
            };
            ExcelFormattingSettings.Add(settingCashA37);
            #endregion

            #region Special handling for CashFlows A45
            ExcelFormattingSetting settingCashA45 = new ExcelFormattingSetting(CASHFLOWS_A45, "A45", "xl/worksheets/sheet4.xml");
            settingCashA45.processFormatting = () =>
            {
                string part = settingCashA45.OriginalText.Substring(0, settingCashA45.OriginalText.Length - 9);
                return string.Format(CASHFLOWS_A45, part.Trim());
            };
            ExcelFormattingSettings.Add(settingCashA45);
            #endregion
        }

        /// <summary>
        /// Format exported Excel file with a temporary path.
        /// </summary>
        /// <param name="tmpPath">A temporary path used for unzip/extract Excel zip file.</param>
        public void FormatExcel(string tmpPath)
        {
            XElement eleSheet = null;
            // load sharedStrings xml document for updating
            XElement eleSharedStrings = XElement.Load(string.Format("{0}/xl/sharedStrings.xml", tmpPath));

            // distinct sheet xml document for searching
            IEnumerable<ExcelFormattingSetting> noduplicates = ExcelFormattingSettings.Distinct(new ExcelFormattingSettingCompare());

            foreach (ExcelFormattingSetting sheet in noduplicates)
            {
                eleSheet = XElement.Load(string.Format("{0}/{1}", tmpPath, sheet.SheetName));
                // update sharedStrings.xml document
                UpdateSharedStringsXMLDoc(sheet.SheetName, eleSharedStrings, eleSheet);
            }

            eleSharedStrings.Save(string.Format("{0}/xl/sharedStrings.xml", tmpPath));
        }

        /// <summary>
        /// Format exported Excel file with a stream.
        /// </summary>
        /// <param name="zipfileStream">A stream of Excel zip file.</param>
        /// <returns>Return a MemoryStream of updated Excel zip file.</returns>
        public Stream FormatExcelWithStream(Stream zipfileStream)
        {
            // copy the current stream to a new stream
            MemoryStream msZip = new MemoryStream();
            long pos = 0;
            pos = zipfileStream.Position;
            zipfileStream.CopyTo(msZip);
            zipfileStream.Position = pos;

            XElement eleSheet = null;

            // load sharedStrings xml document for updating
            XElement eleSharedStrings = null;
            using (Stream mainStream = FindEntryFromZipStream(zipfileStream, "xl/sharedStrings.xml"))
            {
                if (mainStream == null)
                {
                    return msZip;
                }
                eleSharedStrings = XElement.Load(mainStream);
            }

            // distinct sheet xml document for searching
            IEnumerable<ExcelFormattingSetting> noduplicates = ExcelFormattingSettings.Distinct(new ExcelFormattingSettingCompare());
            foreach (ExcelFormattingSetting sheet in noduplicates)
            {
                zipfileStream.Position = 0;
                using (Stream stream = FindEntryFromZipStream(zipfileStream, sheet.SheetName))
                {
                    if (stream != null)
                    {
                        eleSheet = XElement.Load(stream);
                        // update sharedStrings.xml document
                        UpdateSharedStringsXMLDoc(sheet.SheetName, eleSharedStrings, eleSheet);
                    }
                }
            }

            // update to stream
            MemoryStream msEntry = new MemoryStream();
            eleSharedStrings.Save(msEntry);
            // The zipStream is expected to contain the complete zipfile to be updated
            ZipFile zipFile = new ZipFile(msZip);

            zipFile.BeginUpdate();

            // To use the entryStream as a file to be added to the zip,
            // we need to put it into an implementation of IStaticDataSource.
            CustomStaticDataSource sds = new CustomStaticDataSource();
            sds.SetStream(msEntry);

            // If an entry of the same name already exists, it will be overwritten; otherwise added.
            zipFile.Add(sds, "xl/sharedStrings.xml");

            // Both CommitUpdate and Close must be called.
            zipFile.CommitUpdate();
            // Set this so that Close does not close the memorystream
            zipFile.IsStreamOwner = false;
            zipFile.Close();

            return msZip;
        }

        /// <summary>
        /// Find a specific stream with the entry name of the Excel zip file package.
        /// </summary>
        /// <param name="inputStream">The Excel zip file stream.</param>
        /// <param name="entryName">Entry name in the Excel zip file package.</param>
        /// <returns>Return the sepcific stream.</returns>
        private Stream FindEntryFromZipStream(Stream inputStream, string entryName)
        {
            ZipInputStream zipStream = new ZipInputStream(inputStream);
            ZipEntry zippedFile = zipStream.GetNextEntry();

            //Do until no more zipped files left
            while (zippedFile != null)
            {
                byte[] buffer = new byte[2048];
                int bytesRead;

                MemoryStream memoryStream = new MemoryStream();
                // read through the compressed data
                while ((bytesRead = zipStream.Read(buffer, 0, buffer.Length)) != 0)
                {
                    memoryStream.Write(buffer, 0, bytesRead);
                }
                memoryStream.Position = 0;

                if (zippedFile.Name.Equals(entryName))
                {
                    return memoryStream;
                }

                zippedFile = zipStream.GetNextEntry();
            }

            return null;
        }

        /// <summary>
        /// Update formatted strings to the sharedStrings.xml document in Excel zip file.
        /// </summary>
        /// <param name="sheetName"></param>
        /// <param name="navSharedStrings"></param>
        /// <param name="navSheet"></param>
        private void UpdateSharedStringsXMLDoc(string sheetName, XElement eleSharedStrings, XElement eleSheet)
        {
            XNamespace nsSharedStrings = eleSharedStrings.GetDefaultNamespace();
            XNamespace nsSheet = eleSheet.GetDefaultNamespace();
            int i;
            string sContent;

            // update each formatting settings to the sharedStrings xml document
            foreach (ExcelFormattingSetting setting in ExcelFormattingSettings.Where(s => s.SheetName.Equals(sheetName)))
            {
                // find out which si element need to update from the sheet xml document.
                var siIndex = eleSheet.Element(nsSheet + "sheetData")
                                        .Descendants(nsSheet + "c")
                                        .Where(d => d.Attribute("r").Value == setting.ExcelPositionString).FirstOrDefault();
                if (siIndex != null)
                {
                    if (int.TryParse(siIndex.Value, out i))
                    {
                        var siEntry = eleSharedStrings.Elements(nsSharedStrings + "si").ElementAt(i);
                        if (siEntry != null)
                        {
                            var child = siEntry.Element(nsSharedStrings + "t");
                            if (child != null)
                            {
                                setting.OriginalText = child.Value;
                                sContent = setting.ProcessFormatting(setting.processFormatting);
                                // note, cannot set empty content to the new XElement.
                                if (!sContent.Equals(string.Empty))
                                {
                                    XElement newElement = XElement.Parse("<si xmlns=\"http://schemas.openxmlformats.org/spreadsheetml/2006/main\">" + setting.ProcessFormatting(setting.processFormatting) + "</si>");
                                    siEntry.ReplaceWith(newElement);
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    /// <summary>
    /// Excel formatting setting entry class.
    /// </summary>
    public class ExcelFormattingSetting
    {
        private string originalText;
        public string OriginalText
        {
            get { return originalText; }
            set { originalText = value; }
        }

        private string sheetName;
        public string SheetName
        {
            get { return sheetName; }
            set { sheetName = value; }
        }
        
        private string formattingText;
        public string FormattingText
        {
            get { return formattingText; }
            set { formattingText = value; }
        }

        private string excelPositionString;
        public string ExcelPositionString
        {
            get { return excelPositionString; }
            set { excelPositionString = value; }
        }

        public ProcessFormattingDelegate processFormatting;

        public ExcelFormattingSetting()
        {
        }

        public ExcelFormattingSetting(string _formattingtext, string _excelPositionString, string _sheetName)
        {
            this.formattingText = _formattingtext;
            this.excelPositionString = _excelPositionString;
            this.sheetName = _sheetName;
        }

        public string ProcessFormatting(ProcessFormattingDelegate _process)
        {
            if (_process != null)
            {
                return _process();
            }

            // find out all US dollors numbers and put them into a string array.
            Regex reg = new Regex("(\\d+\\,\\d+)|(\\d+)+");
            MatchCollection matches = reg.Matches(this.originalText);
            if (matches.Count > 0)
            {
                string[] arr = new string[matches.Count];
                for (int i = 0; i < matches.Count; i++)
                {
                    arr[i] = matches[i].ToString();
                }
                // return a formatted string with a specified text.
                return string.Format(this.formattingText, arr);
            }

            return this.formattingText;
        }
    }

    public class ExcelFormattingSettingCompare : IEqualityComparer<ExcelFormattingSetting>
    {
        public bool Equals(ExcelFormattingSetting x, ExcelFormattingSetting y)
        {
            //Check whether the compared objects reference the same data.
            if (Object.ReferenceEquals(x, y)) return true;

            //Check whether any of the compared objects is null.
            if (Object.ReferenceEquals(x, null) || Object.ReferenceEquals(y, null))
                return false;

            //Check whether the products' properties are equal.
            return x.SheetName.Equals(y.SheetName);
        }

        public int GetHashCode(ExcelFormattingSetting obj)
        {
            //Check whether the object is null
            if (Object.ReferenceEquals(obj, null)) return 0;

            //Get hash code for the Name field if it is not null.
            return obj.SheetName == null ? 0 : obj.SheetName.GetHashCode();
        }
    }

    public class CustomStaticDataSource : IStaticDataSource
    {
        private Stream _stream;
        // Implement method from IStaticDataSource
        public Stream GetSource()
        {
            return _stream;
        }

        // Call this to provide the memorystream
        public void SetStream(Stream inputStream)
        {
            _stream = inputStream;
            _stream.Position = 0;
        }
    }
}
